/* 2003/2004 Erik Andresen
 * erik@defiant.homedns.org
 *
 * The serial part is copied from miniterm (see the big comment below)
 * Thanks to the Serial Programming HOWTO
 * http://www.tldp.org/HOWTO/Serial-Programming-HOWTO/
 * 
 * You can read more about intelligent interface programming here:
 * http://www.fischertechnik.de/computing/pdf/schnittstelle.pdf
 */


/*
 *  AUTHOR: Sven Goldt (goldt@math.tu-berlin.de)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
*/
/*
 This is like all programs in the Linux Programmer's Guide meant
 as a simple practical demonstration.
 It can be used as a base for a real terminal program.
*/


#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

#define BAUDRATE B9600
#define TRUE 1
#define FALSE 0


typedef struct IInterface {
	int IInt; /* Die Variable die auf das Interface zeigt */
	struct termios newioset; /* Our new settings */
	struct termios saveioset; /* The pointer we save the old settings in*/
	unsigned char MotorStatus; /* hex, Motor state */
	unsigned short int Digitals; /* all Digitals in a hex number */
	unsigned short int AnaX; /* Analog X, dezimal */
	unsigned short AnaY;
	char Digital[8+1]; /* array for digital inputs */
} IInterfaceDef;


/* used before exiting... -just clean up */
void IIexit(IInterfaceDef **myInterface)
{
	tcsetattr((*myInterface)->IInt, TCSANOW, &(*myInterface)->saveioset); /* restore old tty setings */
	/* finally close */
	close((*myInterface)->IInt);
	free(*myInterface);
}


/* Set Motor state, int is 1...4, direction l,r,s for left, right, stop */
void SetMotor(IInterfaceDef **myInterface, int Motor, char direction)
{
	/*unsigned char c = 0xC1;*/
	unsigned char stop=0x3;
	
	switch(direction) {
		case 'l':
			direction=0x1;
			break;
		case 'r':
			direction=0x2;
			break;
		case 's':
			direction=0x0;
			break;
		default:
			printf("Unknown Direction: %c\n", direction);
			direction=0x0;
	}
	/* Reset this Motor to stop */
	stop<<=(Motor-1)*2; /* choose right motor and use OR to target them*/
	(*myInterface)->MotorStatus = stop | (*myInterface)->MotorStatus;
	/* now invert 11 to 00 with a simple XOR */
	(*myInterface)->MotorStatus = stop ^ (*myInterface)->MotorStatus;
	
	direction<<=(Motor-1)*2; /* richtigen Motor whlen */
	(*myInterface)->MotorStatus = direction | (*myInterface)->MotorStatus;
	/* zuerst kommt die Eingangsabfrage */
	/*write(myInterface->IInt, &c, 1);*/
	/* Motoren setzen */
	/*write(myInterface->IInt, &myInterface->MotorStatus, 1);*/
	/* Eingnge lesen */
	/*read(myInterface->IInt, &myInterface->Digitals, 1);*/
}


/* Returns the Motor State */
char GetMotor(IInterfaceDef **myInterface, int Motor)
{
	unsigned char state;
	
	/* Target the bytes, todo it we have to get them to the end...*/
	state = (*myInterface)->MotorStatus<<(4-Motor)*2;
	/* and now to the beginning */
	state>>=6;
	switch(state) {
		case 0x1:
			return 'l';
		case 0x2:
			return 'r';
		default:
			return 's';
	}
}


/* is looking at the dezimal value and converting to a value 1 oder 0 */
int GetDigital(IInterfaceDef *myInterface, int Input)
{
	unsigned char c=1;
	c<<=(Input-1);
	c&=myInterface->Digitals;
	c>>=(Input-1);
	return c;
}


/* Reads AnaX, Digital Input and reset the Motor state */
void ReadX(IInterfaceDef **myInterface)
{
	unsigned char c = 0xC5;
	short int i;
	
	write((*myInterface)->IInt, &c, 1);
	write((*myInterface)->IInt, &(*myInterface)->MotorStatus, 1);
	/* ok, problem: We have to read 3 bytes, but only can get 1 per read.
	 * The first are the Digitals, the last 2 are the Analog, so read the first of the
	 * Analog, do a byte shifting and read the last */
	read((*myInterface)->IInt, &(*myInterface)->Digitals, 1);
	(*myInterface)->Digital[0] = -1;
	for (i=1; i<=8; i++)
		(*myInterface)->Digital[i] = GetDigital((*myInterface), i);
	read((*myInterface)->IInt, &(*myInterface)->AnaX, 1);
	(*myInterface)->AnaX<<=8;
	read((*myInterface)->IInt, &(*myInterface)->AnaX, 1);
}


/* same, but with Y, you can only read X _or_ Y */
void ReadY(IInterfaceDef **myInterface)
{
	unsigned char c = 0xC9;
	short int i;
	
	write((*myInterface)->IInt, &c, 1);
	write((*myInterface)->IInt, &(*myInterface)->MotorStatus, 1);
	/* ok, problem: We have to read 3 bytes, but only can get 1 per read.
	 * The first are the Digitals, the last 2 are the Analog, so read the first of the
	 * Analog, do a byte shifting and read the last */
	read((*myInterface)->IInt, &(*myInterface)->Digitals, 1);
	(*myInterface)->Digital[0] = -1;
	for (i=1; i<=8; i++)
		(*myInterface)->Digital[i] = GetDigital((*myInterface), i);
	read((*myInterface)->IInt, &(*myInterface)->AnaY, 1);
	(*myInterface)->AnaY<<=8;
	read((*myInterface)->IInt, &(*myInterface)->AnaY, 1);
}


/* Only read Digitals, no X, no Y - not used */
void ReadDigitals(IInterfaceDef **myInterface)
{
	unsigned char c = 0xC9;
	short int i;
	
	write((*myInterface)->IInt, &c, 1);
	write((*myInterface)->IInt, &(*myInterface)->MotorStatus, 1);
	read((*myInterface)->IInt, &(*myInterface)->Digitals, 1);
	
	(*myInterface)->Digital[0] = -1;
	for (i=1; i<=8; i++)
		(*myInterface)->Digital[i] = GetDigital((*myInterface), i);
}


/* Just cycle ReadX & Y to make sure the digitals, X, Y and the Motor States are constantly Set */
void II_thread(IInterfaceDef **myInterface)
{
	while(TRUE) {
		ReadX(myInterface);
		ReadY(myInterface);
	}
	
	pthread_exit((void *) 0);
}


/* Init: open the serial console and start the thread */
void IIinit(IInterfaceDef **myInterface	, char *Device)
{	
	pthread_t t1;
	
	if (((*myInterface) = (IInterfaceDef *) malloc (sizeof(IInterfaceDef))) == NULL) {
		fprintf(stderr, "Error getting memory: %s\n", strerror(errno));
		exit(-1);
	}

	/* 
	* Open modem device for reading and writing and not as controlling tty
	* because we don't want to get killed if linenoise sends CTRL-C.
	*/
	/* We use Canonical Input Processing!!! */
	if (((*myInterface)->IInt = open(Device, O_RDWR | O_NOCTTY)) == -1) {
		/* if fail openning device, */
		fprintf(stderr, "Error openning Device:\n");
		/* push the error */
		perror(Device);
		fprintf(stderr, "exiting...");
		/* and exit */
		exit(-1);
	}

	tcgetattr((*myInterface)->IInt, &(*myInterface)->saveioset); /* save current modem settings */
	/* 
	 * BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
	 * CRTSCTS : output hardware flow control (only used if the cable has all necessary lines. See sect. 7 of Serial-HOWTO)
	 * CS8     : 8n1 (8bit,no parity,1 stopbit)
	 * CLOCAL  : local connection, no modem contol
	 * CREAD   : enable receiving characters
	*/
	(*myInterface)->newioset.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD; 

	(*myInterface)->newioset.c_iflag = IGNPAR;
        (*myInterface)->newioset.c_oflag = 0;
        
        /* set input mode (non-canonical, no echo,...) */
        (*myInterface)->newioset.c_lflag = 0;
         
        (*myInterface)->newioset.c_cc[VTIME]    = 0;   /* inter-character timer unused */
        (*myInterface)->newioset.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
	
	/* now clean the modem line and activate the settings for modem */
	tcflush((*myInterface)->IInt, TCIFLUSH);
	tcsetattr((*myInterface)->IInt, TCSANOW, &(*myInterface)->newioset);
		
	if(pthread_create(&t1, NULL, (void *)&II_thread, (IInterfaceDef *)myInterface) != 0)
	{
		fprintf(stderr, "Error in Thread! - %s\n", strerror(errno));
		exit(-1);
	}
}
